-------------------------------------------------------------------------------
-- Package OS_Services_Gnat (body)                                           --
-- (C) Copyright 2000 ADALOG                                                 --
-- Author: J-P. Rosen                                                        --
--                                                                           --
-- Rights to use, distribute or modify this package in any way is hereby     --
-- granted, provided this header is kept unchanged in all versions and the   --
-- associated documentation file is distributed unchanged. Additionnal       --
-- headers or documentation may be added.                                    --
-- All modifications must be properly marked as not originating from Adalog. --
-- If you make a valuable addition, please keep us informed by sending a     --
-- message to rosen.adalog@wanadoo.fr                                        --
--                                                                           --
-- ADALOG is providing training, consultancy and expertise in Ada and        --
-- related software engineering techniques. For more info about our services:--
-- ADALOG                   Tel: +33 1 41 24 31 40                           --
-- 19-21 rue du 8 mai 1945  Fax: +33 1 41 24 07 36                           --
-- 94110 ARCUEIL            E-m: rosen.adalog@wanadoo.fr                     --
--                          URL: http://pro.wanadoo.fr/adalog                --
--                                                                           --
-- This package is distributed in the hope that it will be useful, but WITH- --
-- OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY  --
-- or FITNESS FOR A PARTICULAR PURPOSE.                                      --
-------------------------------------------------------------------------------
with  -- Standard Ada units
  Ada.Exceptions,
  Ada.Strings.Fixed,
  Ada.Characters.Handling,
  Ada.Unchecked_Deallocation;
with -- Gnat specific units
  Gnat.OS_Lib,
  Gnat.Directory_Operations,
  Gnat.Heap_Sort_G;
package body OS_Services_Gnat is

   -- If Forced_OS is set to False, the package will try to recognize the OS.
   -- If your OS is not properly recognized, just set Forced_OS to True,
   -- and initialize Current_OS with the appropriate value.
   Forced_OS  : constant Boolean := False;
   Current_OS : OS_Kind;

   -------------------------
   --                     --
   -- Internal utilities  --
   --                     --
   -------------------------

   Directory_Separator : Character renames Gnat.OS_Lib.Directory_Separator;

   ------------
   -- Adjust --
   ------------

   procedure Adjust (Object : in out File_List) is
   begin
      if Object.Content /= null then
         Object.Content.Ref_Count := Object.Content.Ref_Count + 1;
      end if;
   end Adjust;

   -------------------------
   -- Dictionnary_Compare --
   -------------------------

   -- Compare strings in "dictionnary" order, i.e. ignoring case except to break a tie.

   type Comparison_Order is (Lower, Equal, Greater);

   function Dictionnary_Compare (Left, Right : String) return Comparison_Order is
      use Ada.Characters.Handling;

      Lowered_Left : constant String := To_Lower (Left);
      Lowered_Right: constant String := To_Lower (Right);
   begin
      if Lowered_Left < Lowered_Right then
         return Lower;
      elsif Lowered_Left > Lowered_Right then
         return Greater;
      else    -- use casing to break the tie
         if Left < Right then
            return Lower;
         elsif Left > Right then
            return Greater;
         else
            return Equal;
         end if;
      end if;
   end Dictionnary_Compare;

   ------------
   -- Failed --
   ------------

   -- Every raising of Service_Error is done by this procedure (or the next one) to ensure
   -- an appropriate message is provided
   procedure Failed (Occur : Ada.Exceptions.Exception_Occurrence; Message : String := "") is
      use Ada.Exceptions;
   begin
      if Message = "" then
         Raise_Exception (Service_Error'Identity,
                          Exception_Message (Occur));
      else
         Raise_Exception (Service_Error'Identity,
                          Exception_Message (Occur) & " (" & Message & ')');
      end if;
   end Failed;
   pragma No_Return (Failed);

   ------------
   -- Failed --
   ------------

   procedure Failed (Message : String) is
      use Ada.Exceptions;
   begin
      Raise_Exception (Service_Error'Identity, Message);
   end Failed;
   pragma No_Return (Failed);

   ----------
   -- Free --
   ----------

   procedure Free is new Ada.Unchecked_Deallocation (Info_List, Info_List_Access);

   --------------
   -- Finalize --
   --------------

   procedure Finalize (Object : in out File_List) is
   begin
      if Object.Content /= null then
         Object.Content.Ref_Count := Object.Content.Ref_Count - 1;
         if Object.Content.Ref_Count = 0 then
            Free (Object.Content);
         end if;
      end if;
   end Finalize;

   -----------------
   -- Proper_Name --
   -----------------

   function Proper_Name (Name : String; As_File : Boolean) return String is
      -- Normalizes a file name with appropriate separators
      -- If As_File is true, make sure a trailing separator is removed.
      -- Returned value has lower bound = 1
      Result    : String (1..Name'Length) := Name;
   begin
      -- First, get rid of trivial case:
      if Name = "" then
         return "";
      end if;

      -- Adjust the separators:
      for I in Result'Range loop
         if Result (I) = '/' or Result (I) = '\' then
            Result (I) := Directory_Separator;
         end if;
      end loop;

      -- Optionnaly, remove trailing separator:
      if As_File and Result (Result'Last) = Directory_Separator then
         case Operating_System is
            when Dos_OS =>
               if Result'Length >= 3 and then Result (Result'Last-1) = ':' then
                  -- i.e. "C:\"
                  return Result;
               else
                  return Result (Result'First .. Result'Last-1);
               end if;
            when others =>
               return Result (Result'First .. Result'Last-1);
         end case;
      else
         return Result;
      end if;
   end Proper_Name;

   --------------
   -- Simplify --
   --------------

   function Simplify (Name : String) return String is
      -- Removes any //, /./ and /../ in Name, which must already be absolute
      -- Remove also a trailing /. or /..
      Double_Sep : constant String := Directory_Separator & Directory_Separator;
      This_Dir   : constant String := Directory_Separator & '.';
      Up_Dir     : constant String := Directory_Separator & "..";

      -- Note that we force Result'First = 1 in this declaration:
      -- (would not be the case without the constraint)
      Result   : String (1..Name'Length+2) := Name & "  ";  -- Guard
      Last     : Natural := Name'Length;
      Prev_Sep : Natural := 0;
      Inx      : Natural := 1;
   begin
      while Inx <= Last-1 loop
         if Result (Inx..Inx+1) = Double_Sep then
            -- Remove the second '/'
            Result (Inx+1..Last-1) := Result (Inx+2..Last);
            Last := Last - 1;

         elsif Result (Inx..Inx+1) = This_Dir and
           (Last = Inx+1 or else Result (Inx+2) = Directory_Separator)
         then
            -- Remove the "./"
            Result (Inx+1..Last-2) := Result (Inx+3..Last);
            Last := Last-2;

         elsif Result (Inx..Inx+2) =  Up_Dir and
           (Last = Inx+2 or else Result (Inx+3) = Directory_Separator) then
            -- Look for start of previous directory
            Prev_Sep := Inx-1;
            while Prev_Sep >= 1 and then Result (Prev_Sep) /= Directory_Separator loop
               Prev_Sep := Prev_Sep - 1;
            end loop;

            -- Remove the "/../" and the previous directory
            -- Prev_Sep = 0 should happen only if the name starts with /../
            -- In this case, leave it and let the OS deal with that
            if Prev_Sep /= 0 then
               Result (Prev_Sep+1..Prev_Sep+1 + (Last-(Inx+4))) := Result (Inx+4..Last);
               Last := Last - 4 - (Inx-Prev_Sep-1);
               Inx  := Prev_Sep;
            end if;

         else
            Inx := Inx+1;
         end if;
      end loop;

      return Result (1..Last);
   end Simplify;


   ----------------------------------------------------------------------------------------
   --
   --  General OS Services
   --
   ----------------------------------------------------------------------------------------

   ---------------------
   -- Executable_Path --
   ---------------------

   function Executable_Path return String is
      use Ada.Command_Line, GNAT.OS_Lib;

      function Get_Executable_Name return String is
      begin
         case Operating_System is
            when Dos_OS =>
               -- We know no Ada compiler that generates a .com file...
               return Property (Native_Form, Command_Name, Default => ".exe");
            when others =>
               return Command_Name;
         end case;
      end Get_Executable_Name;

      Exec_Name : constant String := Get_Executable_Name;
   begin
      -- Under DOS-like systems, current directory is searched for the executable
      -- irrespective of path search
      if Operating_System in DOS_Os and then Property (Is_Regular_File, Exec_Name) then
         return Property (Expanded, Exec_Name);
      else
         -- This may seem convoluted, but we need to free the pointer
         -- returned by Locate_Exec_On_Path
         declare
            Name_Ptr :  String_Access  := Locate_Exec_On_Path(Exec_Name);
            Result   : constant String := Name_Ptr.all;
         begin
            Free (Name_Ptr);
            return Result;
         end;
      end if;
   exception
      when Constraint_Error =>
         -- Locate_Exec_On_Path returned null
         Failed ("Unable to locate executable");
  end Executable_Path;

   ---------------------
   -- Get_Environment --
   ---------------------

   function Get_Environment (Name : String) return String is
      use Gnat.OS_Lib;
      Value  : String_Access := GetEnv (Name);
      Result : constant String := Value.all;
   begin
      Free (Value);
      return Result;
   end Get_Environment;

   ----------------------
   -- Operating_System --
   ----------------------

   function Operating_System return OS_Kind is
      pragma Inline (Operating_System);
   begin
      return Current_OS;
   end Operating_System;

   ----------------
   -- OS_Execute --
   ----------------

   procedure OS_Execute (Command  : in     String;
                         Result   :    out Ada.Command_Line.Exit_Status;
                         Blocking : in     Boolean := True) is
      use GNAT.Os_Lib, Ada.Command_Line;

      function Build_Arg_List (Source : String) return Argument_List is
         -- Parse  Command  and  put   each  "word"  in  an  entry  of
         -- Argument_List The  definition of  "word" is taken  from sh
         -- syntax:

         -- A non-quoted backslash  (\) is  the escape  character.  It
         -- preserves  the literal  value of  the next  character that
         -- follows.

         -- Enclosing  characters  in   single  quotes  preserves  the
         -- literal  value  of each  character  within  the quotes.  A
         -- single  quote may  not occur  between single  quotes, even
         -- when preceded by a backslash.

         -- Enclosing  characters  in   double  quotes  preserves  the
         -- literal value  of all  characters within the  quotes, with
         -- the  exception of  $, `,  and \.  The characters  $  and `
         -- retain  their special  meaning within  double  quotes. The
         -- backslash retains  its special meaning  only when followed
         -- by  one  of the  following  characters:  $,  `, ",  \,  or
         -- <newline>.  A double  quote  may be  quoted within  double
         -- quotes by preceding it with a backslash.

         Start, Stop          : Natural;
         New_Start            : Natural;
         Within_Single_Quotes : Boolean := False;
         Within_Double_Quotes : Boolean := False;
      begin
         if Source = "" then
            return (1..0 => null);
         end if;

         Start := Source'First;
         while Start <= Source'Last and then Source (Start) <= ' ' loop
            Start := Start + 1;
         end loop;

         Stop := Start;
         while Stop <= Source'Last loop
            if Within_Single_Quotes then
               if Source (Stop) = ''' then
                  Within_Single_Quotes := False;
               end if;
            elsif Within_Double_Quotes then
               case Source (Stop) is
                  when '"' =>
                     Within_Double_Quotes := False;
                  when ''' =>
                     Within_Single_Quotes := True;
                  when '\' =>
                     if Stop /= Source'Last then
                        Stop := Stop + 1;
                     end if;
                  when others =>
                     null;
               end case;
            else
               case Source (Stop) is
                  when Character'First .. ' ' =>
                     exit;
                  when '"' =>
                     Within_Double_Quotes := True;
                  when ''' =>
                     Within_Single_Quotes := True;
                  when others =>
                     null;
               end case;
            end if;
            Stop := Stop + 1;
         end loop;

         if Stop = Start then  -- Source contained only spaces
              return (1..0 => null);
         end if;

         New_Start := Stop;
         if Operating_System = Unix then
            -- Surrounding quotes must not be included in parameter
            if Source (Start) = ''' or else Source (Start) = '"' then
               Start := Start+1;
            end if;
            if Source (Stop-1) = ''' or else Source (Stop-1) = '"' then
               Stop := Stop-1;
            end if;
         end if;

         return (1 => new String'(Source (Start .. Stop-1))) &
           Build_Arg_List (Source (New_Start .. Source'Last));
      end Build_Arg_List;

      Args : Argument_List := Build_Arg_List (Command);
      -- Args (1) is the command name

      Extra_Arg     : String_Access;
      Has_Succeeded : Boolean;
   begin
      -- Unix (and presumably other OSes) require the full path name
      -- to the command.
      -- On DOS, it would have been nice to force execution in a
      -- separate "COMMAND.COM /C", to save the user to worry about
      -- internal/external commands. However we had to step back,
      -- because the return code is not transmitted by COMMAND.COM,
      -- and every command would appear as succesful. Sigh...
      -- Of course, users can use Shell_Execute if they want to execute
      -- in a subshell.
      case Operating_System is
         when Dos_OS =>
            if Blocking then
               Spawn (Args(1).all, Args(2..Args'Last), Has_Succeeded);
            else
               Has_Succeeded := Non_Blocking_Spawn (Args(1).all, Args(2..Args'Last)) /= Invalid_Pid;
            end if;

         when others =>
            Extra_Arg := Locate_Exec_On_Path (Args (1).all);
            if Extra_Arg = null then
               Has_Succeeded := False;
            else
               if Blocking then
                  Spawn (Extra_Arg.all, Args (2..Args'Last), Has_Succeeded);
               else
                  Has_Succeeded := Non_Blocking_Spawn (Extra_Arg.all, Args (2..Args'Last)) /= Invalid_Pid;
               end if;
           end if;
      end case;

      if Has_Succeeded then
         Result := Success;
      else
         Result := Failure;
      end if;

      -- Release resources
      Free (Extra_Arg); -- OK if null
      for I in Args'Range loop
         Free(Args(I));
      end loop;

   end OS_Execute;

   -------------------
   -- Shell_Execute --
   -------------------

   procedure Shell_Execute (Command  : in     String;
                            Result   :    out Ada.Command_Line.Exit_Status;
                            Blocking : in     Boolean := True) is
      -- Executes Command in a sub-shell
   begin
      case Operating_System is
         when Dos =>
            OS_Execute ("COMMAND /C " & Command, Result, Blocking);
         when NT | OS_2 =>
            OS_Execute ("CMD /C " & Command, Result, Blocking);
         when Unix =>
            OS_Execute ("sh -c '" & Command & ''', Result, Blocking);
         when Unknown =>
            Failed ("Unimplemented");
      end case;
   end Shell_Execute;


   --------------------------------------------
   --                                        --
   -- File name pattern matching utilitities --
   --                                        --
   --------------------------------------------

   -------------
   -- Compile --
   -------------

   function Compile (Pattern : String; Ignore_Case : Boolean := Default_Ignore_Case) return Compiled_Pattern is
      use Gnat.Regexp;
   begin
      -- Call the inherited "Compile":
      return Compile (Pattern, Glob => True, Case_Sensitive => not Ignore_Case);
   exception
      when Error_In_Regexp =>
         Failed ("Illegal pattern: " & Pattern);
   end Compile;

   ------------------------------
   -- Match (Compiled_Pattern) --
   ------------------------------

   function Match (Name : String; Pattern : Compiled_Pattern) return Boolean is
      -- Since we declared a conformant "Match" in the spec, it is NOT inherited
      -- from Regexp. Call the one in Regexp with appropriate type conversions.
      use Gnat.Regexp;
   begin
      return Match (Name, Regexp (Pattern));
   end Match;

   --------------------
   -- Match (String) --
   --------------------

   function Match (Name : String; Pattern : String; Ignore_Case : Boolean := Default_Ignore_Case) return Boolean is
      Compiled_Pat : constant Compiled_Pattern := Compile (Pattern, Ignore_Case => Ignore_Case);
   begin
      return Match (Name, Compiled_Pat);
   end Match;



   -----------------------------------------
   --                                     --
   --  File name and properties functions --
   --                                     --
   -----------------------------------------


   ---------
   -- "+" --
   ---------
   function "+" (Left, Right : String_Property) return String_Property is
      -- cannot use rename as body because "or" is intrinsic
      pragma Inline ("+");
   begin
      return Left or Right;
   end "+";

   ----------
   -- Join --
   ----------

   function Join (Path : String; Name : String) return String is
   begin
      if Path = "" then
         return Name;
      end if;

      if Path (Path'Last) = Directory_Separator then
         return Path & Name;
      else
         return Path & Directory_Separator & Name;
      end if;
   end Join;

   ----------------
   -- Properties --
   ----------------

   function Properties (Name : String) return File_Properties is
      use Ada.Strings.Unbounded;

      Result : File_Properties (Property (File_Kind, Name));
   begin
      Result.Name := To_Unbounded_String (Property (OS_Services_Gnat.Name+Extension, Name));
      Result.Path := To_Unbounded_String (Property (Expanded+Device+Path,            Name));
      case Result.Kind is
         when File | Directory =>
            Result.Change_Time := Property (Change_Time, Name);
            Result.Size        := Property (Size,        Name);
         when others =>
            null;
      end case;
      return Result;
   end Properties;

   -----------------------
   -- Property (string) --
   -----------------------

   function Property (Query : String_Property; Name : String; Default : String := "") return String is
      -- Analyze full names. Syntax:
      -- DOS : device:path\name.ext
      -- Unix: /path/name.ext

      -- Under DOS, Current_Dir includes the current drive
      -- Def_Path_First indicates the first char of the (default) path; characters before
      -- Def_Path_First (if any) belong to the device name.
      -- TBSL: Does not work for path that are VolumeRelative
      Def_Path       : constant String := Current_Dir & Directory_Separator;
      Def_Path_First : Natural := Def_Path'First;

      Good_Query   : String_Property := Query;
      -- Note that Good_Name'First = 1 (guaranteed by Proper_Name)
      Good_Name    : constant String := Proper_Name (Name,    As_File => Query (As_File_Flag));
      Good_Default : constant String := Proper_Name (Default, As_File => False);

      type Descriptor is
        record
           Device_First : Natural;
           Device_Last  : Natural := 0;
           Path_First   : Natural;
           Path_Last    : Natural := 0;
           Name_First   : Natural;
           Name_Last    : Natural := 0;
           Ext_First    : Natural;
           Ext_Last     : Natural := 0;
        end record;

      Name_Descr    : Descriptor;
      Default_Descr : Descriptor;

      function Parse_Name (Source : String) return Descriptor is
         -- Initializes the XXX_First and XXX_Last for each XXX part of Source
         -- if XXX is not present, XXX_Last = 0 and XXX_First >= 1
         use Ada.Strings, Ada.Strings.Fixed;

         Temp   : Natural;
         Result : Descriptor;
      begin
         -- Device
         case Operating_System is
            when Dos_OS =>
               if Source'Length >= 2 and then Source (2) = ':' then
                  Result.Device_Last := 2;
               end if;
            when others =>
               -- No notion of device
               null;
         end case;

         -- Extension
         Temp := Index (Source, ".", Going => Backward);
         if Temp /= 0 then
            Result.Ext_Last  := Source'Last;
            Result.Name_Last := Temp - 1;
         else
            Result.Name_Last := Source'Last;
         end if;

         -- Path; adjust extension if necessary
         Temp := Index (Source, (1 => Directory_Separator), Going => Backward);
         if Temp /= 0 then
            Result.Path_Last := Temp;
            if Result.Ext_Last /= 0 and Result.Name_Last < Result.Path_Last then
               -- the '.' was in the path...
               Result.Ext_Last  := 0;
               Result.Name_Last := Source'Last;
            end if;
         end if;

         -- Adjust XXX_First pointers according to missing elements
         Result.Device_First := 1;
         if Result.Device_Last = 0 then
            Result.Path_First := 1;
         else
            Result.Path_First := Result.Device_Last + 1;
         end if;
         if Result.Path_Last = 0 then
            Result.Name_First := Result.Path_First;
         else
            Result.Name_First := Result.Path_Last + 1;
         end if;
         if Result.Name_Last = 0 then
            Result.Ext_First := Result.Name_First;
         else
            Result.Ext_First := Result.Name_Last + 1;
         end if;

         return Result;
      end Parse_Name;

      function PartD return String is
         -- Computes Device part
      begin
         if not Good_Query (Device_Flag) then
            return "";

         elsif Name_Descr.Device_First <= Name_Descr.Device_Last then
            return Good_Name (Name_Descr.Device_First..Name_Descr.Device_Last);

         elsif Default_Descr.Device_First <= Default_Descr.Device_Last then
            -- Part not provided, default provided
            return Good_Default (Default_Descr.Device_First..Default_Descr.Device_Last);

         else
            -- Part not provided, no default
            if Good_Query (Expanded_Flag) then
               return Def_Path (Def_Path'First..Def_Path_First-1);
            else
               return "";
            end if;
         end if;
      end PartD;

      function PartP return String is
         -- Computes Path part
      begin
         if not Good_Query (Path_Flag) then
            return "";

         elsif Name_Descr.Path_First <= Name_Descr.Path_Last then
            if Good_Query (Expanded_Flag) and
              Good_Name (Name_Descr.Path_First) /= Directory_Separator -- Expanded and relative
            then
               return Join (Def_Path (Def_Path_First..Def_Path'Last),
                            Good_Name (Name_Descr.Path_First..Name_Descr.Path_Last));
            else  -- Not expanded, or already absolute
               return Good_Name (Name_Descr.Path_First..Name_Descr.Path_Last);
            end if;

         elsif Default_Descr.Path_First <= Default_Descr.Path_Last then
            -- Part not provided, default provided
            if Good_Query (Expanded_Flag) and
              Good_Default (Default_Descr.Path_First) /= Directory_Separator -- Expanded and relative
            then
               return Join (Def_Path (Def_Path_First..Def_Path'Last),
                            Good_Default (Default_Descr.Path_First..Default_Descr.Path_Last));
            else  -- Not expanded, or already absolute
               return Good_Default (Default_Descr.Path_First..Default_Descr.Path_Last);
            end if;

         else
            -- Part not provided, no default
            if Good_Query (Expanded_Flag) then
               return Def_Path (Def_Path_First..Def_Path'Last);
            else
               return "";
            end if;
         end if;
      end PartP;

      function PartN return String is
         -- Computes Name part
      begin
         if not Good_Query (Name_Flag) then
            return "";

         elsif Name_Descr.Name_First <= Name_Descr.Name_Last then
            return Good_Name (Name_Descr.Name_First..Name_Descr.Name_Last);

         else
            -- Part not provided; Ok if Def_From > Def_To: return ""
            return Good_Default (Default_Descr.Name_First..Default_Descr.Name_Last);
         end if;
      end PartN;

      function PartE return String is
         -- Computes Extension part
      begin
         if not Good_Query (Ext_Flag) then
            return "";

         elsif Name_Descr.Ext_First <= Name_Descr.Ext_Last then
            return Good_Name (Name_Descr.Ext_First..Name_Descr.Ext_Last);

         else
            -- Part not provided; Ok if Def_From > Def_To: return ""
            return Good_Default (Default_Descr.Ext_First..Default_Descr.Ext_Last);
         end if;
      end PartE;

   begin -- Property
      if Good_Query (Device_Flag..Version_Flag) = (Device_Flag..Version_Flag => False) then
         -- no part specified <=> take all parts
         Good_Query (Device_Flag..Version_Flag) := (Device_Flag..Version_Flag => True);
      end if;

      case Operating_System is
         when Dos_OS =>
            -- Current_Dir under DOS-like systems includes the device
            Def_Path_First := Def_Path'First + 2;
         when others =>
            -- No notion of device for other OSes.
            Def_Path_First := Def_Path'First;
      end case;

      Name_Descr    := Parse_Name (Good_Name);
      Default_Descr := Parse_Name (Good_Default);

      if Good_Query (Expanded_Flag) then
         return Simplify (PartD & PartP & PartN & PartE);
      else
         return PartD & PartP & PartN & PartE;
      end if;
   end Property;

   ------------------------
   -- Property (Boolean) --
   ------------------------

   function Property (Query : Boolean_Property; Name : String) return Boolean is
      Good_Name : constant String := Proper_Name (Name, As_File => True);
   begin
      case Query is
         when Exists =>
            return Property (Is_Regular_File, Good_Name) or else Property (Is_Directory, Good_Name);
         when Is_Directory =>
            return Gnat.OS_Lib.Is_Directory (Good_Name);
         when Is_Regular_File =>
            return Gnat.OS_Lib.Is_Regular_File (Good_Name);
      end case;
   end Property;

   ---------------------
   -- Property (Time) --
   ---------------------

   function Property (Query : Time_Property; Name : String) return Ada.Calendar.Time
   is
      use Gnat.OS_Lib, Ada.Calendar;
      Date   : OS_Time;
      Year   : Year_Type;
      Month  : Month_Type;
      Day    : Day_Type;
      Hour   : Hour_Type;
      Minute : Minute_Type;
      Second : Second_Type;
   begin
      case Query is
         when Change_Time =>
            Date := File_Time_Stamp (Proper_Name (Name, As_File => True));
      end case;

      GM_Split (Date, Year, Month, Day, Hour, Minute, Second);
      return Time_Of (Year, Month, Day, Duration (3600*Hour + 60*Minute + Second));
   end Property;

   -----------------------------
   -- Property (Long_Integer) --
   -----------------------------

   function Property (Query : Long_Property; Name : String) return Long_Integer is
      use GNAT.OS_Lib;
      Good_Name : constant String := Proper_Name (Name, As_File => True);
      C_Name    : constant String := Good_Name & Ascii.Nul;
      FD        : File_Descriptor;
      Result    : Long_Integer;
   begin
      case Query is
         when Size =>
            if Property (Is_Directory, Good_Name) then
               return 0;
            end if;

            FD := Open_Read (C_Name (C_Name'First)'Address, Binary);
            if FD = Invalid_FD then
               Failed ("Size of non-existent file: " & Name);
            end if;
            Result := File_Length (FD);
            Close (FD);
            return Result;
      end case;
   end Property;

   ----------------------------
   -- Property (Enum_Values) --
   ----------------------------

   function Property (Query : Enum_Property; Name : String) return Enum_Values is
      Good_Name : constant String := Proper_Name (Name, As_File => True);
   begin
      case Query is
         when File_Kind =>
            -- This is the best we can do for the moment...
            if Property (Is_Directory, Good_Name) then
               return Directory;
            else
               return File;
            end if;
         when Path_Kind =>
            -- Check if name absolute
            if Good_Name (Good_Name'First) = Directory_Separator or else
              (Operating_System in Dos_OS  and then
              Good_Name'Length >= 3 and then
              Good_Name(Good_Name'First+1 .. Good_Name'First+2) = ':' & Directory_Separator)
            then
               return Absolute;
            end if;

            -- Name is relative here.
            -- Check for an explicit drive:
            if Operating_System in Dos_OS and then
              Good_Name'Length >= 2 and then
              Good_Name (Good_Name'First+1) = ':'
            then
               return VolumeRelative;
            end if;

            -- Normal case (Name is relative to current drive)
            return Relative;
      end case;
   end Property;

   -----------------------------------
   --                               --
   --  Directory iterator functions --
   --                               --
   -----------------------------------

   ----------
   -- "+" --
   ----------

   function "+" (Left, Right : File_List) return File_List is
      use Ada.Strings.Unbounded, GNAT.OS_Lib;
      Out_Info : Info_Table (1..Left.Content.Length + Right.Content.Length); --Maximum possible size
   begin
      -- Get rid of trivial cases
      if Length (Left) = 0 then
         return Right;
      elsif Length (Right) = 0 then
         return Left;
      end if;

      -- Get rid of unsorted case
      if not Left.Content.Sorted or not Right.Content.Sorted then
         Out_Info := Left.Content.Infos & Right.Content.Infos;

         -- Adjust the character pointers in the (former) Right:
         for I in Left.Content.Length + 1 .. Left.Content.Length + Right.Content.Length loop
            Out_Info (I).Path_Start := Out_Info (I).Path_Start + Length (Left.All_Names) + 1;
            Out_Info (I).Sep_Pos    := Out_Info (I).Sep_Pos    + Length (Left.All_Names) + 1;
            Out_Info (I).Name_End   := Out_Info (I).Name_End   + Length (Left.All_Names) + 1;
         end loop;

         return (Ada.Finalization.Controlled with
                 All_Names =>
                   Left.All_Names & Path_Separator & Right.All_Names,
                 Content =>
                   new Info_List'(Length    => Left.Content.Length + Right.Content.Length,
                                  Ref_Count => 1,
                                  Sorted    => False,
                                  Infos     => Out_Info)
                 );
      end if;

      -- Here we are in the case of sorted "+"
      declare
         Left_Info  : Info_Table renames Left.Content.Infos;
         Right_Info : Info_Table renames Right.Content.Infos;
         Left_Inx   : Natural := Left_Info'First;
         Right_Inx  : Natural := Right_Info'First;
         Out_Inx    : Natural := 0;
         All_Names  : Unbounded_String;

         procedure Add (Elem : File_List; Inx : Natural) is
            Start, Stop : Natural;
         begin
            Out_Inx            := Out_Inx + 1;
            Out_Info (Out_Inx) := Elem.Content.Infos (Inx);
            if All_Names = Null_Unbounded_String then
               All_Names := To_Unbounded_String (Slice (Elem.All_Names,
                                                        Elem.Content.Infos (Inx).Path_Start,
                                                        Elem.Content.Infos (Inx).Name_End));
               Start := 1;
               Stop  := Length (All_Names);
            else
               Start := Length (All_Names) + 2;
               All_Names :=
                 All_Names &
                 Path_Separator &
                 To_Unbounded_String (Slice (Elem.All_Names,
                                             Elem.Content.Infos (Inx).Path_Start,
                                             Elem.Content.Infos (Inx).Name_End));
               Stop := Length (All_Names);
            end if;
            Out_Info (Out_Inx).Path_Start := Start;
            Out_Info (Out_Inx).Name_End   := Stop;
            Out_Info (Out_Inx).Sep_Pos    :=
              Elem.Content.Infos (Inx).Sep_Pos
              - Elem.Content.Infos (Inx).Path_Start
              + Start;
         end Add;

      begin
         while Left_Inx <= Left_Info'Last and Right_Inx <= Right_Info'Last loop
            case Dictionnary_Compare (Slice (Left.All_Names,
                                             Left_Info (Left_Inx).Path_Start,
                                             Left_Info (Left_Inx).Sep_Pos - 1),
                                      Slice (Right.All_Names,
                                             Right_Info (Right_Inx).Path_Start,
                                             Right_Info (Right_Inx).Sep_Pos - 1)) is
            when Lower =>
               Add (Left, Left_Inx);
               Left_Inx := Left_Inx + 1;
            when Greater =>
               Add (Right, Right_Inx);
               Right_Inx := Right_Inx + 1;
            when Equal =>
               case Dictionnary_Compare (Slice (Left.All_Names,
                                                Left_Info (Left_Inx).Sep_Pos + 1,
                                                Left_Info (Left_Inx).Name_End),
                                         Slice (Right.All_Names,
                                                Right_Info (Right_Inx).Sep_Pos + 1,
                                                Right_Info (Right_Inx).Name_End)) is
                  when Lower =>
                     Add (Left, Left_Inx);
                     Left_Inx := Left_Inx + 1;
                  when Greater =>
                     Add (Right, Right_Inx);
                     Right_Inx := Right_Inx + 1;
                  when Equal => -- two identical entries
                     Add (Right, Right_Inx);
                     Right_Inx := Right_Inx + 1;
                     Left_Inx  := Left_Inx + 1;
               end case;
            end case;
         end loop;

         -- Add the non-exhausted side.
         for I in Left_Inx .. Left_Info'Last loop
            Add (Left, I);
         end loop;

         for I in Right_Inx .. Right_Info'Last loop
            Add (Right, I);
         end loop;

         return (Ada.Finalization.Controlled with
                 All_Names =>
                   All_Names,
                 Content =>
                   new Info_List'(Length    => Out_Inx,
                                  Ref_Count => 1,
                                  Sorted    => True,
                                  Infos     => Out_Info (1..Out_Inx)
                                  )
            );
      end;
   end "+";

   ----------------
   -- Build_List --
   ----------------

   function Build_List (Pattern : in String  := "*";
                        Sort    : in Boolean := False)
     return File_List is

      function Build_One_List (Pattern : in String;
                               Sort    : in Boolean)
        return File_List is
         use Gnat.Directory_Operations, Gnat.OS_Lib, Ada.Strings.Unbounded;

         File_Pattern : constant Compiled_Pattern := Compile (Pattern     => Property (Name+Extension, Pattern),
                                                              Ignore_Case => Operating_System /= Unix);
         Path_String : constant String := Property (Expanded+Device+Path, Pattern);
         Dir         : Dir_Type;
         Buffer      : String (1..512); -- Long enough for any file name
         Last        : Natural;
         File_Num    : Natural := 0;
         Result      : File_List;
         Start       : Positive;
         Midle       : Positive;
         Stop        : Positive;

         -- Instantiation of generic heap sort:
         Temp : Inner_Info;

         function LT (OP1, OP2 : Natural) return Boolean is
         begin
            if OP1 = 0 then
               return Dictionnary_Compare (Slice (Result.All_Names, Temp.Path_Start, Temp.Name_End),
                                           Property (Expanded, Result, OP2)) = Lower;
            elsif OP2 = 0 then
               return Dictionnary_Compare (Property (Expanded, Result, OP1),
                                           Slice (Result.All_Names, Temp.Path_Start, Temp.Name_End)) = Lower;
            else
               return Dictionnary_Compare (Property (Expanded, Result, OP1),
                                           Property (Expanded, Result, OP2)) = Lower;
            end if;
         end LT;

         procedure Move (From : Natural; To : Natural) is
         begin
            if From = 0 then
               Result.Content.Infos (To) := Temp;
            elsif To = 0 then
               Temp := Result.Content.Infos (From);
            else
               Result.Content.Infos (To) := Result.Content.Infos (From);
            end if;
         end Move;

         package File_List_Sort is new Gnat.Heap_Sort_G (Move, LT);

      begin  -- Build_One_List
         if Path_String = "" then
            Open (Dir, Current_Dir);
         elsif Path_String = "\" or else
           (Path_String'Length = 3 and then Path_String (Path_String'Last-1..Path_String'Last) = ":\") then
            Open (Dir, Path_String);
         else
            Open (Dir, Path_String);
         end if;

         loop
            Read (Dir, Buffer, Last);
            exit when Last = 0;

            if Buffer (1..Last) /= "." and
              Buffer (1..Last) /= ".." and
              Match (Buffer (1..Last), File_Pattern)
            then
               if Result.All_Names = Null_Unbounded_String then
                  Result.All_Names := To_Unbounded_String (Property (Expanded, Path_String & Buffer (1..Last)));
               else
                  Result.All_Names :=
                    Result.All_Names &
                    Path_Separator &
                    Property (Expanded, Path_String & Buffer (1..Last));
               end if;
               File_Num := File_Num + 1;
            end if;
         end loop;

         Result.Content := new Info_List (Length => File_Num);
         Result.Content.Ref_Count := 1;
         -- We cannot consider that an empty list or a singleton is always sorted,
         -- since it would force sorting on the "+" operation
         Result.Content.Sorted := Sort;

         Start := 1;
         for I in 1 .. File_Num loop
            Midle := Start;  -- Guard for the case of root files ("\name")
            Stop  := Start+1;
            while Element (Result.All_Names, Stop) /= Path_Separator loop
               if Element (Result.All_Names, Stop) = Directory_Separator then
                  Midle := Stop;
               end if;
               Stop := Stop + 1;
               exit when Stop > Length (Result.All_Names); -- Case of the last name
            end loop;

            -- We can't use an aggregate to initialize the infos, since some fields
            -- must be left uninitialized due to lazy evaluation.
            declare
               Elem : Inner_Info renames Result.Content.Infos (I);
            begin
               Elem.Path_Start  := Start;
               Elem.Sep_Pos     := Midle;
               Elem.Name_End    := Stop-1;
               Elem.Kind_Valid  := False;
               Elem.Mtime_Valid := False;
               Elem.Size_Valid  := False;
            end;

            Start := Stop+1;
         end loop;

         if Sort then
            File_List_Sort.Sort (Length (Result));
         end if;

         return Result;

      exception
         when Occur : Directory_Error =>
            Failed (Occur, "Illegal path: " & Path_String);
      end Build_One_List;

      In_Quotes : Boolean := False;
      Inx_Pat   : Natural := Pattern'First;
      Clean_Pat : String (1..Pattern'Length);
      Inx_Clean : Natural := 0;
      use Ada.Finalization, Ada.Strings.Unbounded;
   begin  -- Build_List
      -- Skip leading spaces and control characters
      while Inx_Pat <= Pattern'Last and then  Pattern (Inx_Pat) <= ' ' loop
         Inx_Pat := Inx_Pat + 1;
      end loop;

      if Inx_Pat > Pattern'Last then
         -- Pattern contained only spaces or control character (or was "")
         -- Return empty list
         return (Controlled with Null_Unbounded_String, null);
      end if;

      loop
         if Pattern (Inx_Pat) = '"' then
            In_Quotes := not In_Quotes;
         elsif not In_Quotes and Pattern (Inx_Pat) <= ' ' then
            exit;
         else
            Inx_Clean := Inx_Clean + 1;
            Clean_Pat (Inx_Clean) := Pattern (Inx_Pat);
         end if;
         exit when Inx_Pat >= Pattern'Last;
         Inx_Pat := Inx_Pat + 1;
      end loop;

      if Inx_Pat >= Pattern'Last then
         return Build_One_List (Clean_Pat (1 .. Inx_Clean), Sort);
      else
         return
           Build_One_List (Clean_Pat (1 .. Inx_Clean), Sort) +
           Build_List     (Pattern (Inx_Pat+1 .. Pattern'Last),  Sort);
      end if;
   end Build_List;

   --------------------
   -- Generic_Filter --
   --------------------

   function Generic_Filter (List : File_List) return File_List is
      use Ada.Finalization, Ada.Strings.Unbounded, Gnat.OS_Lib;

      Local_Table : Info_Table (1..List.Content.Infos'Length);
      Local_Inx   : Natural := 0;
      All_Names   : Unbounded_String;

   begin
      for I in 1..List.Content.Length loop
         if Is_Selected (List, I) then
            Local_Inx := Local_Inx+1;
            Local_Table (Local_Inx) := List.Content.Infos (I);
            if All_Names = Null_Unbounded_String then
               Local_Table (Local_Inx).Path_Start := 1;
               All_Names := To_Unbounded_String (Property (Expanded, List, I));
            else
               Local_Table (Local_Inx).Path_Start := Length (All_Names) + 2;
               All_Names := All_Names & Path_Separator & Property (Expanded, List, I);
            end if;
            Local_Table (Local_Inx).Name_End := Length (All_Names);
            Local_Table (Local_Inx).Sep_Pos  :=
              List.Content.Infos (I).Sep_Pos
              - List.Content.Infos (I).Path_Start
              + Local_Table (Local_Inx).Path_Start;
         end if;
      end loop;

      return (Controlled with
              All_Names =>
                All_Names,
              Content =>
                new Info_List'(Length    => Local_Inx,
                               Ref_Count => 1,
                               Sorted    => List.Content.Sorted,
                               Infos     => Local_Table (1..Local_Inx)));
   end Generic_Filter;

   ------------
   -- Length --
   ------------

   function Length (List : File_List) return Natural is
   begin
      if List.Content = null then
         return 0;
      else
         return List.Content.Length;
      end if;
   end Length;

   ----------------
   -- Properties --
   ----------------

   function  Properties (List : File_List; Index : Positive) return File_Properties is
      use Ada.Strings.Unbounded;
      -- Always access properties through the Property operations (not directly from the
      -- record), to make sure that the lazy evaluation is performed if necessary.
   begin
      declare
         Result : File_Properties (Property (File_Kind, List, Index));
      begin
         Result.Name := To_Unbounded_String (Property (Name+Extension,       List, Index));
         Result.Path := To_Unbounded_String (Property (Expanded+Device+Path, List, Index));
         if Result.Kind in File..Directory then
            Result.Change_Time := Property (Change_Time, List, Index);
            Result.Size        := Property (Size,        List, Index);
         end if;
         return Result;
      end;
   end Properties;

   -----------------------
   -- Property (String) --
   -----------------------

   function Property (Query : String_Property; List : File_List; Index : Positive) return String is
      -- Note that there is no lazy evaluation of name properties.
      use Ada.Strings.Unbounded, Gnat.OS_Lib;
      Elem : Inner_Info renames List.Content.Infos(Index);
   begin
      return Property (Query, Slice (List.All_Names, Elem.Path_Start, Elem.Name_End));
   end Property;

   ----------------------------
   -- Property (Enum_Values) --
   ----------------------------

   function Property (Query : Iter_Enum_Property; List : File_List; Index : Positive) return Enum_Values is
      -- Path_Type does not make sense, since we always provide the Path and Name.
      Elem : Inner_Info renames List.Content.Infos(Index);
   begin
      case Query is
         when File_Kind =>
            if not Elem.Kind_Valid then
               Elem.Kind       := Property (File_Kind, Property (Expanded, List, Index));
               Elem.Kind_Valid := True;
            end if;
            return Elem.Kind;
      end case;
   end Property;

   ------------------------
   -- Property (Boolean) --
   ------------------------

   function Property (Query : Boolean_Property; List : File_List; Index : Positive) return Boolean is
      Elem : Inner_Info renames List.Content.Infos(Index);
   begin
      if not Elem.Kind_Valid then
         Elem.Kind       := Property (File_Kind, Property (Expanded, List, Index));
         Elem.Kind_Valid := True;
      end if;

      case Query is
         when Is_Directory =>
            return Elem.Kind = Directory;
         when Is_Regular_File =>
            return Elem.Kind = File;
          when Exists =>
            return Property (Query, Property (Expanded, List, Index));
     end case;
   end Property;

   ---------------------
   -- Property (Time) --
   ---------------------

   function Property (Query : Time_Property;   List : File_List; Index : Positive) return Ada.Calendar.Time is
      Elem : Inner_Info renames List.Content.Infos(Index);
   begin
     case Query is
         when Change_Time =>
            if not Elem.Mtime_Valid then
               Elem.Mtime       := Property (Change_Time, Property (Expanded, List, Index));
               Elem.Mtime_Valid := True;
            end if;
            return Elem.Mtime;
      end case;
   end Property;

   -----------------------------
   -- Property (Long_Integer) --
   -----------------------------

   function Property (Query : Long_Property;   List : File_List; Index : Positive) return Long_Integer is
      Elem : Inner_Info renames List.Content.Infos(Index);
   begin
      case Query is
         when Size =>
            if not Elem.Size_Valid then
               Elem.Size       := Property (Size, Property (Expanded, List, Index));
               Elem.Size_Valid := True;
            end if;
            return Elem.Size;
      end case;
   end Property;

   ---------------
   -- To_String --
   ---------------

   function To_String (List : File_List; Separator : Character := ' ') return String is
      -- Note: there were several possibilities for writing this function.
      -- The one we chose has the benefit that the string is scanned only once as long
      -- as file names do not contain the Separator.

      use Gnat.OS_Lib, Ada.Strings.Unbounded;

      -- Worst case: all names must be enclosed in quotes:
      Result     : String (1..Length (List.All_Names) + 2*Length (List));
      Original   : constant String := To_String (List.All_Names);
      Start_Inx  : Positive := 1;
      Out_Inx    : Natural  := 0;
      In_Quotes  : Boolean  := False;
   begin
      for In_Inx in Original'Range loop
         if Original (In_Inx) = Path_Separator then
            if In_Quotes then
               -- Close the quote
               Out_Inx := Out_Inx + 1;
               Result (Out_Inx) := '"';
            end if;

            Out_Inx := Out_Inx + 1;
            Result (Out_Inx) := Separator;

            Start_Inx := Out_Inx+1;
            In_Quotes := False;

         elsif Original (In_Inx) = Separator and not In_Quotes then
            -- OOPS! name must be in quotes
            Result (Start_Inx+1..Out_Inx+1) := Result (Start_Inx..Out_Inx);
            Result (Start_Inx)              := '"';
            Out_Inx          := Out_Inx + 1 + 1;
            Result (Out_Inx) := Separator;
            In_Quotes        := True;

         else
            Out_Inx := Out_Inx + 1;
            Result (Out_Inx) := Original (In_Inx);
         end if;
      end loop;

      if In_Quotes then
         -- Last file was quoted, close the quote
         Out_Inx := Out_Inx + 1;
         Result (Out_Inx) := '"';
      end if;

      return Result(1..Out_Inx);
   end To_String;


   ------------------------------------
   --                                --
   --  File and directory operations --
   --                                --
   ------------------------------------

   ----------------
   -- Change_Dir --
   ----------------

   procedure Change_Dir (To : String) is
   begin
      Gnat.Directory_Operations.Change_Dir (To);
   end Change_Dir;

   ---------------
   -- Copy_File --
   ---------------

   procedure Copy_File (Pattern : String; To : String) is
      use Ada.Command_Line, GNAT.OS_Lib;

      List    : File_List;
      Result  : Exit_Status;
      Last_To : Natural;
   begin
      case Operating_System is
         when Dos_OS =>
            -- Dos copy does not accept a target directory with a trailing '\',
            -- except when it's the root (another sigh)...
            if To (To'Last) = Directory_Separator and then
              To'Last /= To'First and then
              To (To'Last-1) /= ':'
            then
               Last_To := To'Last - 1;
            else
               Last_To := To'Last;
            end if;

            List := Build_List (Pattern);
            if Length (List) > 1 and then not Property (Is_Directory, To) then
               -- Copy several files to single file => error
               Failed ("Copy multiple files to single target");
            end if;

            for I in 1 .. Length (List) loop
               if Property (Is_Directory, List, I) then
                  -- Source is directory
                  if Property (Is_Directory, To) then
                     declare
                        Target : constant String := Join (To, Property (Name+Extension, List, I));
                     begin
                        if not Property (Exists, Target) then
                           Create_Dir (Target);
                        elsif not Property (Is_Directory, Target) then
                           Failed ("Copy directory to existing non-directory");
                        end if;
                        Copy_File ('"' & Join (Property (Expanded, List, I), "*") & '"', Target);
                     end;
                  elsif Property (Exists, To) then
                     -- Target exists and is not a directory =>
                     -- don't overwrite it
                     Failed ("Copy directory to existing non-directory");
                  else
                     -- Target does not exist
                     Create_Dir (To);
                     Copy_File ('"' & Join (Property (Expanded, List, I), "*") & '"', To);
                  end if;
               else
                  -- Source is regular file

                  -- Add quotes around the name for the case of spaces in file names
                  -- YAS (Yet Another Sigh)...
                  Shell_Execute ("copy "
                                 & '"' & Property (Native_Form, List, I) & '"'
                                 & ' '
                                 & '"' & Property (Native_Form, To (To'First .. Last_To)) & '"'
                                 & " >nul:",
                                 Result);
                  if Result /= Success then
                     Failed ("Copy command failed");
                  end if;
               end if;
            end loop;

         when Unix =>
            OS_Execute ("cp -p -r "
                        & Property (Native_Form, Pattern)
                        & ' '
                        & '"' & Property (Native_Form, To) & '"',
                        Result);
            if Result /= Success then
               Failed ("Copy command failed");
            end if;

         when Unknown =>
            Failed ("Not implemented");
      end case;
   end Copy_File;

   ----------------
   -- Create_Dir --
   ----------------

   procedure Create_Dir (Name : String) is
      use Ada.Command_Line;
      Result : Exit_Status;
   begin
      case Operating_System is
         when Dos_OS =>
            Shell_Execute ("mkdir """ & Property (Native_Form, Name) & '"', Result);
         when Unix =>
            OS_Execute ("mkdir """ & Property (Native_Form, Name) & '"', Result);
         when Unknown =>
            Failed ("Unimplemented");
      end case;

      if Result /= Success then
         Failed ("Mkdir command failed");
      end if;
   end Create_Dir;

   -----------------
   -- Current_Dir --
   -----------------

   function Current_Dir return String is
      use Gnat.Directory_Operations;
      Result : constant String := Get_Current_Dir;
   begin
      -- We want the Current_dir as a real file name
      -- => if the string returned by Gnat includes a last
      --    directory separator, get rid of it.
      if Result (Result'Last) = Directory_Separator then
         return Result (Result'First .. Result'Last - 1);
      else
         return Result;
      end if;
   end Current_Dir;

   -----------------
   -- Delete_File --
   -----------------

   procedure Delete_File (Pattern : String) is
      use Gnat.OS_Lib, Ada.Command_Line;
      Result : Exit_Status := Success;
      List   : File_List;
   begin
      case Operating_System is
         when Dos_OS =>
            -- We cannot give the Pattern directly to the dos del command, because our syntax
            -- for wildcarding is not the same as Dos.

            List := Build_List (Pattern);
            for I in 1 .. Length (List) loop
               if Property (Is_Directory, List, I) then
                  Delete_File ('"' & Join (Property (Expanded, List, I), "*") & '"');
                  Shell_Execute ("rmdir " & '"' & Property (Expanded, List, I) & '"', Result);
               else
                  Shell_Execute ("del "
                                 & '"' & Property (Expanded, List, I) & '"'
                                 & " >nul:",
                                 Result);
               end if;
            end loop;

         when Unix =>
            -- Must use Shell_Execute and not OS_Execute for wildcards to be effective
            Shell_Execute ("rm -rf "
                        & Property (Native_Form, Pattern),
                        Result);

         when Unknown =>
            Failed ("Unimplemented");
      end case;

      if Result /= Success then
         Failed ("Del command failed");
      end if;
   end Delete_File;

   ---------------
   -- Move_File --
   ---------------

   procedure Move_File (Pattern : String; To : String) is
      use Ada.Command_Line;
      Result : Exit_Status;
   begin
      case Operating_System is
         when Dos_OS =>
            Copy_File (Pattern, To);
            -- Copy_File raises Service_Error if not succesful
            -- Let it propagate

            Delete_File (Pattern);
         when Unix =>
            OS_Execute ("mv "
                        & Property (Native_Form, Pattern)
                        & ' '
                        & '"' & Property (Native_Form, To) & '"',
                        Result);
            if Result /= Success then
               Failed ("Mv command failed");
            end if;
         when Unknown =>
            Failed ("Not implemented");
      end case;
   end Move_File;


   -----------------
   -- Rename_File --
   -----------------

   procedure Rename_File (File : String; To : String) is
      use Ada.Command_Line;
      Result : Exit_Status;
   begin
      case Operating_System is
         when Dos_OS =>
            -- Ren does not return an error code if the file exists,
            -- so we must test it by hand
            if Property (Exists, To) then
               Failed ("Target exists for rename");
            end if;
            Shell_Execute ("ren "
                           & '"' & Property (Native_Form, File) & '"'
                           & ' '
                           & '"' & Property (Native_Form, To) & '"'
                           & " >nul:",
                           Result);
         when Unix =>
            OS_Execute ("mv "
                        & '"' & Property (Native_Form, File) & '"'
                        & ' '
                        & '"' & Property (Native_Form, To) & '"',
                        Result);
         when Unknown =>
            Failed ("Not implemented");
      end case;

      if Result /= Success then
         Failed ("Ren/mv command failed");
      end if;
   end Rename_File;


   use Gnat.OS_Lib;

begin

   if not Forced_OS then
      -- Try to guess the OS.
      -- Note that GNAT is particularly unhelpful, since System_Name is always
      -- SYSTEM_NAME_GNAT, irrespectively of the actual system.
      -- How do we recognize VMS, MacOS and OS/2?
      -- Note that the trick below will recognize OS/2 as Dos (or
      -- NT?), and that it might even work...
      case Directory_Separator is
         when '/' =>
            Current_OS := Unix;
         when '\' =>
            if Locate_Exec_On_Path ("cmd.exe") /= null then
               Current_OS := NT;
            else
               Current_OS := Dos;
            end if;
         when others =>
            Current_OS := Unknown;
      end case;
   end if;

   Default_Ignore_Case := Current_OS /= Unix;
end OS_Services_Gnat;
