-------------------------------------------------------------------------------
-- Package OS_Services_TCL (body)                                            --
-- (C) Copyright 2000 ADALOG                                                 --
-- Author: J-P. Rosen                                                        --
--                                                                           --
-- Rights to use, distribute or modify this package in any way is hereby     --
-- granted, provided this header is kept unchanged in all versions and the   --
-- associated documentation file is distributed unchanged. Additionnal       --
-- headers or documentation may be added.                                    --
-- All modifications must be properly marked as not originating from Adalog. --
-- If you make a valuable addition, please keep us informed by sending a     --
-- message to rosen.adalog@wanadoo.fr                                        --
--                                                                           --
-- ADALOG is providing training, consultancy and expertise in Ada and        --
-- related software engineering techniques. For more info about our services:--
-- ADALOG                   Tel: +33 1 41 24 31 40                           --
-- 19-21 rue du 8 mai 1945  Fax: +33 1 41 24 07 36                           --
-- 94110 ARCUEIL            E-m: rosen.adalog@wanadoo.fr                     --
--                          URL: http://pro.wanadoo.fr/adalog                --
--                                                                           --
-- This package is distributed in the hope that it will be useful, but WITH- --
-- OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY  --
-- or FITNESS FOR A PARTICULAR PURPOSE.                                      --
-------------------------------------------------------------------------------
with  -- Standard Ada units
  Ada.Text_IO,
  Ada.Exceptions,
  Ada.Characters.Handling,
  Ada.Unchecked_Deallocation,
  Ada.Strings.Fixed,
  Interfaces.C.Strings;
with  -- Tash units
  Tcl.Ada,
  Chelper;
with  -- Reusable components
  Calendar_Utilities;
package body OS_Services_TCL is

   -----------------
   --             --
   -- Global data --
   --             --
   -----------------


   -- If Forced_OS is set to False, the package will try to recognize the OS.
   -- If your OS is not properly recognized, just set Forced_OS to True,
   -- and initialize Current_OS with the appropriate value.
   Forced_OS  : constant Boolean := False;
   Current_OS : OS_Kind;

   Directory_Separator : Character; -- Initialized in body of package

   --  This is one, global, Tcl interpreter, shared by all
   --  subprograms in this package, except those marked "reentrant".
   --  This is the reason why the features not marked "reentrant" should
   --  never be called by two tasks at the same time; it may produce strange
   --  results, including, but not limited to, memory violations.
   Interp : Tcl.Tcl_Interp := Tcl.Tcl_Createinterp;


   ----------------------------------------------------------
   --                                                      --
   -- Finalization, RefCount and dynamic memory management --
   --                                                      --
   ----------------------------------------------------------

   ------------
   -- At_End --
   ------------

   -- A finalizable type and object to make sure Tcl_Finalize is called when the program
   -- terminates.
   -- The object must be declared at the beginning of the module, to make sure that its
   -- Finalized is called LAST, i.e. when nobody wants to use TCL any more.
   type At_End is new Ada.Finalization.Limited_Controlled with null record;
   procedure Finalize (Item : in out At_End);
   procedure Finalize (Item : in out At_End) is
   begin
      Tcl.Tcl_Finalize;
   end Finalize;
   At_End_Object : At_End;

   ------------
   -- Adjust --
   ------------

   procedure Adjust (Object : in out File_List) is
   begin
      if Object.Content /= null then
         Object.Content.Ref_Count := Object.Content.Ref_Count + 1;
      end if;
   end Adjust;

   --------------
   -- Finalize --
   --------------

   procedure Finalize (Object : in out File_List) is
      procedure Free is new Ada.Unchecked_Deallocation (Info_List, Info_List_Access);
   begin
      if Object.Content /= null then
         Object.Content.Ref_Count := Object.Content.Ref_Count - 1;
         if Object.Content.Ref_Count = 0 then
            Free (Object.Content);
         end if;
      end if;
   end Finalize;


   ----------------------------------
   --                              --
   --  Internal utilities and data --
   --                              --
   ----------------------------------

   -- Forward declarations, to keep subprograms alphabetically sorted:
   procedure Failed (Message : String);
   function Protected_Name (Source : String) return String;

   ------------------
   -- Check_Result --
   ------------------

   procedure Check_Result (Code : Interfaces.C.Int) is
      use Tcl, Tcl.Ada, Interfaces.C;
   begin
      if Code = Tcl_Error then
         Failed (Tcl_Getvar(Interp, "errorInfo", 0));
      end if;
   end Check_Result;

   -------------------------
   -- Dictionnary_Compare --
   -------------------------

   -- Compare strings in "dictionnary" order, i.e. ignoring case except to break a tie.
   -- This function must match the comparison order used by TCL to build the list.

   type Comparison_Order is (Lower, Equal, Greater);
   function Dictionnary_Compare (Left, Right : Ada.Strings.Unbounded.Unbounded_String) return Comparison_Order is
      use Ada.Characters.Handling, Ada.Strings.Unbounded;

      Lowered_Left : constant String := To_Lower (To_String(Left));
      Lowered_Right: constant String := To_Lower (To_String(Right));
   begin
      if Lowered_Left < Lowered_Right then
         return Lower;
      elsif Lowered_Left > Lowered_Right then
         return Greater;
      else    -- use casing to break the tie
         if Left < Right then
            return Lower;
         elsif Left > Right then
            return Greater;
         else
            return Equal;
         end if;
      end if;
   end Dictionnary_Compare;

   -------------------
   -- Eval (String) --
   -------------------

   -- Evaluates a TCL expr given as String, result String

   function Eval (Program : String) return String is
      use Tcl.Ada;
   begin
      Tcl_Eval (Interp, Program);
      return Tcl_Getstringresult(Interp);
   end Eval;

   ----------------
   -- Eval (obj) --
   ----------------

   -- Evaluates a TCL expr given as String, result Obj

   function Eval (Program : String) return Tcl.Tcl_Obj is
      use Tcl, Tcl.Ada;
   begin
      Tcl_Eval (Interp, Program);
      return Tcl_Getobjresult(Interp);
   end Eval;

   ------------
   -- Failed --
   ------------

   -- Every raising of Service_Error is done by this procedure (or the next one) to ensure
   -- an appropriate message is provided
   procedure Failed (Occur : Ada.Exceptions.Exception_Occurrence; Message : String := "") is
      use Ada.Exceptions;
   begin
      if Message = "" then
         Raise_Exception (Service_Error'Identity,
                          Exception_Message (Occur));
      else
         Raise_Exception (Service_Error'Identity,
                          Exception_Message (Occur) & " (" & Message & ')');
      end if;
   end Failed;
   pragma No_Return (Failed);

   ------------
   -- Failed --
   ------------

   procedure Failed (Message : String) is
      use Ada.Exceptions;
   begin
      Raise_Exception (Service_Error'Identity, Message);
   end Failed;
   pragma No_Return (Failed);

   -----------------
   -- Proper_Name --
   -----------------

   function Proper_Name (Name : String; As_File : Boolean) return String is
      -- Normalizes a file name with appropriate separators
      -- If As_File is true, make sure a trailing separator is removed.
      -- Returned value has lower bound = 1
      Result    : String (1..Name'Length) := Name;
   begin
      -- First, get rid of trivial case:
      if Name = "" then
         return "";
      end if;

      -- Adjust the separators:
      for I in Result'Range loop
         if Result (I) = '/' or Result (I) = '\' then
            Result (I) := Directory_Separator;
         end if;
      end loop;

      -- Optionnaly, remove trailing separator:
      if As_File and Result (Result'Last) = Directory_Separator then
         case Operating_System is
            when Dos_OS =>
               if Result'Length >= 3 and then Result (Result'Last-1) = ':' then
                  -- i.e. "C:\"
                  return Result;
               else
                  return Result (Result'First .. Result'Last-1);
               end if;
            when others =>
               return Result (Result'First .. Result'Last-1);
         end case;
      else
         return Result;
      end if;
   end Proper_Name;

   --------------------
   -- Protected_List --
   --------------------

   -- Take a list of elements separated by spaces, and surrounds each element with {}
   -- An element can include spaces, if surrounded by double quotes

   function Protected_List (Source : String) return String is
      Start, Stop   : Natural;
      New_Start     : Natural;
      Within_Quotes : Boolean := False;
   begin
      if Source = "" then
         return "";
      end if;

      Start := Source'First;
      while Start <= Source'Last and then Source (Start) <= ' ' loop
         Start := Start + 1;
      end loop;

      Stop := Start;
      while Stop <= Source'Last loop
         if Within_Quotes then
            if Source (Stop) = '"' then
               Within_Quotes := False;
            end if;
         else
            case Source (Stop) is
               when Character'First .. ' ' =>
                  exit;
               when '"' =>
                  Within_Quotes := True;
               when others =>
                  null;
            end case;
         end if;
         Stop := Stop + 1;
      end loop;

      if Stop = Start then  -- Source contained only spaces
           return "";
      end if;

      New_Start := Stop;

      -- Surrounding quotes (if any) must not be included in parameter
      if Source (Start) = '"' then
         Start := Start+1;
      end if;
      if Source (Stop-1) = '"' then
         Stop := Stop-1;
      end if;

      return Protected_Name (Source (Start .. Stop-1)) & ' ' &
             Protected_List (Source (New_Start .. Source'Last));
   end Protected_List;

   --------------------
   -- Protected_Name --
   --------------------

   -- Surrounds Source with {} so it forms a single word for Tcl
   -- However, there is a problem if the name ends with a '\', since this would quote
   -- the closing '}'. Since this is used to pass patterns, we can call Proper_Name on
   -- the source, with As_File => True.
   -- There is still a nasty case however: in Dos, if the name is of the form "C:\",
   -- Proper_Name will NOT remove the '\' (this is necessary in Dos to designate the root).
   -- But since there can be no space in the name in this case, we do not enclose it in {},
   -- but double the '\' instead.
   -- You are free to send your opinion about file name handling in Dos to Mr. Bill Gates.

   function Protected_Name (Source : String) return String is
      Local_Name : constant String := Proper_Name (Source, As_File => True);
   begin
      if Local_Name = "" then
         return "{}";

      elsif Local_Name (Local_Name'Last) /= '\' then
         return '{' & Local_Name & '}';

      else
         -- The nasty case
         return Local_Name & '\';
      end if;
   end Protected_Name;

   ----------------
   -- Protect_BS --
   ----------------

   -- Double any '\' in Input to prevent it from being interpreted
   -- as quotation character by Tcl

   function Protect_BS (Input : String) return String is
      use Ada.Strings.Fixed;
      Result : String (1..Input'Length + Ada.Strings.Fixed.Count (Input, "\"));
      Inx    : Positive := Result'First;
   begin
      for I in Input'Range loop
         if Input (I) = '\' then
            Result (Inx) := '\';
            Inx := Inx+1;
         end if;
         Result (Inx) := Input (I);
         Inx := Inx + 1;
      end loop;

      return Result;
   end Protect_BS;

   --------------------
   -- Protected_Eval --
   --------------------

   -- Make an eval with a local interpreter
   -- to be safe in case of multi-tasking

   procedure Protected_Eval (Command : String) is
      use Tcl, Tcl.Ada;
      Local_Interp : Tcl_Interp := Tcl_Createinterp;
   begin
      Tcl_Eval (Local_Interp, Command);
      Tcl_Deleteinterp (Local_Interp);
   exception
      when others =>
         Tcl_Deleteinterp (Local_Interp);
         raise;
   end Protected_Eval;

   --------------
   -- Simplify --
   --------------

   function Simplify (Name : String) return String is
      -- Removes any //, /./ and /../ in Name, which must already be absolute
      -- Remove also a trailing /. or /..
      Double_Sep : constant String := Directory_Separator & Directory_Separator;
      This_Dir   : constant String := Directory_Separator & '.';
      Up_Dir     : constant String := Directory_Separator & "..";

      -- Note that we force Result'First = 1 in this declaration:
      -- (would not be the case without the constraint)
      Result   : String (1..Name'Length+2) := Name & "  ";  -- Guard
      Last     : Natural := Name'Length;
      Prev_Sep : Natural := 0;
      Inx      : Natural := 1;
   begin
      while Inx <= Last-1 loop
         if Result (Inx..Inx+1) = Double_Sep then
            -- Remove the second '/'
            Result (Inx+1..Last-1) := Result (Inx+2..Last);
            Last := Last - 1;

         elsif Result (Inx..Inx+1) = This_Dir and
           (Last = Inx+1 or else Result (Inx+2) = Directory_Separator)
         then
            -- Remove the "./"
            Result (Inx+1..Last-2) := Result (Inx+3..Last);
            Last := Last-2;

         elsif Result (Inx..Inx+2) =  Up_Dir and
           (Last = Inx+2 or else Result (Inx+3) = Directory_Separator) then
            -- Look for start of previous directory
            Prev_Sep := Inx-1;
            while Prev_Sep >= 1 and then Result (Prev_Sep) /= Directory_Separator loop
               Prev_Sep := Prev_Sep - 1;
            end loop;

            -- Remove the "/../" and the previous directory
            -- Prev_Sep = 0 should happen only if the name starts with /../
            -- In this case, leave it and let the OS deal with that
            if Prev_Sep /= 0 then
               Result (Prev_Sep+1..Prev_Sep+1 + (Last-(Inx+4))) := Result (Inx+4..Last);
               Last := Last - 4 - (Inx-Prev_Sep-1);
               Inx  := Prev_Sep;
            end if;

         else
            Inx := Inx+1;
         end if;
      end loop;

      return Result (1..Last);
   end Simplify;


   --------------------------
   -- Tcl_GetStringFromObj --
   --------------------------

   -- This one is a procedure in Tcl.Ada, but it is more
   -- convenient to have it as a function:

   function Tcl_GetStringFromObj (Objptr : in Tcl.Tcl_Obj) return String is
      Junk : aliased Interfaces.C.Int;
   begin -- Tcl_GetStringFromObj
      return Chelper.Value (Tcl.Tcl_Getstringfromobj (
            Objptr,
            Junk'access));
   end Tcl_GetStringFromObj;


   --------------------------
   --                      --
   --  General OS Services --
   --                      --
   --------------------------

   ---------------------
   -- Executable_Path --
   ---------------------

   function Executable_Path return String is
      use Ada.Command_Line, Interfaces.C.Strings, Tcl.Ada;

      function Get_Executable_Name return String is
      begin
         case Operating_System is
            when Dos_OS =>
               -- We know no Ada compiler that generates a .com file...
               return Property (Native_Form, Command_Name, Default => ".exe");
            when others =>
               return Command_Name;
         end case;
      end Get_Executable_Name;

      Exec_Name : constant String := Get_Executable_Name;
   begin
      -- Under DOS-like systems, current directory is searched for the executable
      -- irrespective of path search
      if Operating_System in DOS_Os and then Property (Is_Regular_File, Exec_Name) then
         return Property (Expanded, Exec_Name);
      else
         return Value (Tcl_GetNameOfExecutable);
      end if;
   end Executable_Path;

   ---------------------
   -- Get_Environment --
   ---------------------

   function Get_Environment (Name : String) return String is
      use TCL.Ada;
   begin
      return Eval ("set env(" & Name & ')');
   exception
      when TCL_Error_Exception =>
         -- Variable not foud
         return "";
   end Get_Environment;

   ----------------------
   -- Operating_System --
   ----------------------

   function Operating_System return OS_Kind is
      pragma Inline (Operating_System);
   begin
      return Current_OS;
   end Operating_System;

   ----------------
   -- OS_Execute --
   ----------------

   procedure OS_Execute (Command  : in     String;
                         Result   :    out Ada.Command_Line.Exit_Status;
                         Blocking : in     Boolean := True) is
      use Ada.Command_Line, TCL.Ada, Ada.Text_IO;

      -- We must substitute all '\' as '\\' to prevent TCL from removing them
      Patched_Command : constant String := Protect_BS (Command);
   begin
      if Blocking then
         declare
            -- The result of Eval is the standard output of the command
            Output : constant String := Eval ("exec -- " & Patched_Command);
         begin
            Put_Line (Output);
            Result := Success;
         exception
            when TCL_Error_Exception =>
               Put_Line (Output);
               Result := Failure;
         end;
      else
         -- Standard output really goes to standard output in this case
         Tcl_Eval (Interp, "exec -- " & Patched_Command & " &");
      end if;
   end OS_Execute;

   -------------------
   -- Shell_Execute --
   -------------------

   procedure Shell_Execute (Command  : in     String;
                            Result   :    out Ada.Command_Line.Exit_Status;
                            Blocking : in     Boolean := True) is
      -- Executes Command in a sub-shell
   begin
      case Operating_System is
         when Dos =>
            OS_Execute ("COMMAND /C " & Command, Result, Blocking);
         when NT | OS_2 =>
            OS_Execute ("CMD /C " & Command, Result, Blocking);
         when Unix =>
            OS_Execute ("sh " & Command, Result, Blocking);
         when MacOS | Unknown =>
            Failed ("Unimplemented");
      end case;
   end Shell_Execute;


   --------------------------------------------
   --                                        --
   -- File name pattern matching utilitities --
   -- (Glob-style matching)                  --
   --                                        --
   --------------------------------------------

   -------------
   -- Compile --
   -------------

   function Compile (Pattern : String; Ignore_Case : Boolean := Default_Ignore_Case) return Compiled_Pattern is
      use Ada.Strings.Unbounded;
   begin
      return (To_Unbounded_String (Pattern), Ignore_Case);
   end Compile;

   -----------
   -- Match --
   -----------

   function Match (Name : String; Pattern : String; Ignore_Case : Boolean := Default_Ignore_Case) return Boolean is
      use Tcl.Ada;
      use type Interfaces.C.Int;
   begin
      if Ignore_Case then
         return Tcl_StringCaseMatch (Name, Pattern, 1) = 1;
      else
         return Tcl_StringMatch (Name, Pattern) = 1;
      end if;
   end Match;

   -----------
   -- Match --
   -----------

   function Match   (Name : String; Pattern : Compiled_Pattern) return Boolean is
      use Ada.Strings.Unbounded;
   begin
      return Match (Name, To_String (Pattern.Pattern), Pattern.Ignore_Case);
   end Match;


   ----------------------------------------------------------------------------------------
   --
   --  File name and properties functions
   --
   ----------------------------------------------------------------------------------------

   ---------
   -- "+" --
   ---------
   function "+" (Left, Right : String_Property) return String_Property is
      -- cannot use rename as body because "or" is intrinsic
      pragma Inline ("+");
   begin
      return Left or Right;
   end "+";

   ----------
   -- Join --
   ----------

   function Join (Path : String; Name : String) return String is
      -- Normally, a TCL-purist implementation should use "file join"
      -- Unfortunately, Tcl Join has the unfortunate habit of changing
      -- '\' into '/' under Dos; of course, we could patch it back,
      -- but it is so simpler to simply do the job ourself...
   begin
      if Path = "" then
         return Name;
      end if;

      if Path (Path'Last) = Directory_Separator then
         return Path & Name;
      else
         return Path & Directory_Separator & Name;
      end if;
   end Join;

   ----------------
   -- Properties --
   ----------------

   function Properties (Name : String) return File_Properties is
      use Ada.Strings.Unbounded;

      Result : File_Properties (Property (File_Kind, Name));
   begin
      Result.Name := To_Unbounded_String(Property(Os_Services_Tcl.Name+Extension, Property (Expanded, Name)));
      Result.Path := To_Unbounded_String(Property(Path,                           Property (Expanded, Name)));
      case Result.Kind is
         when File | Directory =>
            Result.Change_Time := Property (Change_Time, Name);
            Result.Access_Time := Property (Access_Time, Name);
            Result.Size        := Property (Size, Name);
         when others =>
            null;
      end case;
      return Result;
   end Properties;

   -----------------------
   -- Property (string) --
   -----------------------

   function Property (Query : String_Property; Name : String; Default : String := "") return String is
      -- We chose to analyze names manually rather than relying on the corresponding
      -- TCL functions, because the Tcl "file xxx" commands were to far away from what we
      -- needed. And if we start manually analyzing strings, why not make the whole job directly...

      -- Analyze full names. Syntax:
      -- DOS : device:path\name.ext
      -- Unix: /path/name.ext

      -- Under DOS, Current_Dir includes the current drive
      -- Def_Path_First indicates the first char of the (default) path; characters before
      -- Def_Path_First (if any) belong to the device name.
      -- TBSL: Does not work for path that are VolumeRelative
      Def_Path       : constant String := Current_Dir & Directory_Separator;
      Def_Path_First : Natural := Def_Path'First;

      Good_Query   : String_Property := Query;
      -- Note that Good_Name'First = 1 (guaranteed by Proper_Name)
      Good_Name    : constant String := Proper_Name (Name,    As_File => Query (As_File_Flag));
      Good_Default : constant String := Proper_Name (Default, As_File => False);

      type Descriptor is
        record
           Device_First : Natural;
           Device_Last  : Natural := 0;
           Path_First   : Natural;
           Path_Last    : Natural := 0;
           Name_First   : Natural;
           Name_Last    : Natural := 0;
           Ext_First    : Natural;
           Ext_Last     : Natural := 0;
        end record;

      Name_Descr    : Descriptor;
      Default_Descr : Descriptor;

      function Parse_Name (Source : String) return Descriptor is
         -- Initializes the XXX_First and XXX_Last for each XXX part of Source
         -- if XXX is not present, XXX_Last = 0 and XXX_First >= 1
         use Ada.Strings, Ada.Strings.Fixed;

         Temp   : Natural;
         Result : Descriptor;
      begin
         -- Device
         case Operating_System is
            when Dos_OS =>
               if Source'Length >= 2 and then Source (2) = ':' then
                  Result.Device_Last := 2;
               end if;
            when others =>
               -- No notion of device
               null;
         end case;

         -- Extension
         Temp := Index (Source, ".", Going => Backward);
         if Temp /= 0 then
            Result.Ext_Last  := Source'Last;
            Result.Name_Last := Temp - 1;
         else
            Result.Name_Last := Source'Last;
         end if;

         -- Path; adjust extension if necessary
         Temp := Index (Source, (1 => Directory_Separator), Going => Backward);
         if Temp /= 0 then
            Result.Path_Last := Temp;
            if Result.Ext_Last /= 0 and Result.Name_Last < Result.Path_Last then
               -- the '.' was in the path...
               Result.Ext_Last  := 0;
               Result.Name_Last := Source'Last;
            end if;
         end if;

         -- Adjust XXX_First pointers according to missing elements
         Result.Device_First := 1;
         if Result.Device_Last = 0 then
            Result.Path_First := 1;
         else
            Result.Path_First := Result.Device_Last + 1;
         end if;
         if Result.Path_Last = 0 then
            Result.Name_First := Result.Path_First;
         else
            Result.Name_First := Result.Path_Last + 1;
         end if;
         if Result.Name_Last = 0 then
            Result.Ext_First := Result.Name_First;
         else
            Result.Ext_First := Result.Name_Last + 1;
         end if;

         return Result;
      end Parse_Name;

      function PartD return String is
         -- Computes Device part
      begin
         if not Good_Query (Device_Flag) then
            return "";

         elsif Name_Descr.Device_First <= Name_Descr.Device_Last then
            return Good_Name (Name_Descr.Device_First..Name_Descr.Device_Last);

         elsif Default_Descr.Device_First <= Default_Descr.Device_Last then
            -- Part not provided, default provided
            return Good_Default (Default_Descr.Device_First..Default_Descr.Device_Last);

         else
            -- Part not provided, no default
            if Good_Query (Expanded_Flag) then
               return Def_Path (Def_Path'First..Def_Path_First-1);
            else
               return "";
            end if;
         end if;
      end PartD;

      function PartP return String is
         -- Computes Path part
      begin
         if not Good_Query (Path_Flag) then
            return "";

         elsif Name_Descr.Path_First <= Name_Descr.Path_Last then
            if Good_Query (Expanded_Flag) and
              Good_Name (Name_Descr.Path_First) /= Directory_Separator -- Expanded and relative
            then
               return Join (Def_Path (Def_Path_First..Def_Path'Last),
                            Good_Name (Name_Descr.Path_First..Name_Descr.Path_Last));
            else  -- Not expanded, or already absolute
               return Good_Name (Name_Descr.Path_First..Name_Descr.Path_Last);
            end if;

         elsif Default_Descr.Path_First <= Default_Descr.Path_Last then
            -- Part not provided, default provided
            if Good_Query (Expanded_Flag) and
              Good_Default (Default_Descr.Path_First) /= Directory_Separator -- Expanded and relative
            then
               return Join (Def_Path (Def_Path_First..Def_Path'Last),
                            Good_Default (Default_Descr.Path_First..Default_Descr.Path_Last));
            else  -- Not expanded, or already absolute
               return Good_Default (Default_Descr.Path_First..Default_Descr.Path_Last);
            end if;

         else
            -- Part not provided, no default
            if Good_Query (Expanded_Flag) then
               return Def_Path (Def_Path_First..Def_Path'Last);
            else
               return "";
            end if;
         end if;
      end PartP;

      function PartN return String is
         -- Computes Name part
      begin
         if not Good_Query (Name_Flag) then
            return "";

         elsif Name_Descr.Name_First <= Name_Descr.Name_Last then
            return Good_Name (Name_Descr.Name_First..Name_Descr.Name_Last);

         else
            -- Part not provided; Ok if Def_From > Def_To: return ""
            return Good_Default (Default_Descr.Name_First..Default_Descr.Name_Last);
         end if;
      end PartN;

      function PartE return String is
         -- Computes Extension part
      begin
         if not Good_Query (Ext_Flag) then
            return "";

         elsif Name_Descr.Ext_First <= Name_Descr.Ext_Last then
            return Good_Name (Name_Descr.Ext_First..Name_Descr.Ext_Last);

         else
            -- Part not provided; Ok if Def_From > Def_To: return ""
            return Good_Default (Default_Descr.Ext_First..Default_Descr.Ext_Last);
         end if;
      end PartE;

   begin -- Property
      if Good_Query (Device_Flag..Version_Flag) = (Device_Flag..Version_Flag => False) then
         -- no part specified <=> take all parts
         Good_Query (Device_Flag..Version_Flag) := (Device_Flag..Version_Flag => True);
      end if;

      case Operating_System is
         when Dos_OS =>
            -- Current_Dir under DOS-like systems includes the device
            Def_Path_First := Def_Path'First + 2;
         when others =>
            -- No notion of device for other OSes.
            Def_Path_First := Def_Path'First;
      end case;

      Name_Descr    := Parse_Name (Good_Name);
      Default_Descr := Parse_Name (Good_Default);

      if Good_Query (Expanded_Flag) then
         return Simplify (PartD & PartP & PartN & PartE);
      else
         return PartD & PartP & PartN & PartE;
      end if;
   end Property;

   ------------------------
   -- Property (Boolean) --
   ------------------------

   Tcl_Bool_Info_Names : array (Boolean_Property) of String (1..11) :=
     (Exists          => "exists     ",
      Is_Directory    => "isdirectory",
      Is_Regular_File => "isfile     ",
      Is_Owned        => "owned      ",
      Is_Readable     => "readable   ",
      Is_Writable     => "writable   ",
      Is_Executable   => "executable ");

   function Property (Query : Boolean_Property; Name : String) return Boolean is
      use Ada.Characters.Handling, Tcl.Ada;
   begin
      return Eval ("file " &
                   Tcl_Bool_Info_Names(Query) &
                   ' ' & Protected_Name (Name)) = "1";
   exception
      when Occur : Tcl_Error_Exception =>
         Failed (Occur);
   end Property;

   ---------------------
   -- Property (Time) --
   ---------------------

   Tcl_Time_Info_Names : array (Time_Property) of String (1..6) := (
      Access_Time => "atime ",
      Change_Time => "mtime ");

   function Property (Query : Time_Property; Name : String) return Ada.Calendar.Time is
      use Ada.Characters.Handling;
      use Calendar_Utilities, Tcl.Ada;
   begin
      return Value (Eval ("clock format [file " & Tcl_Time_Info_Names(Query) &
                          '{' & Protected_Name (Name) & "] -format {%d/%m/%Y %H:%M:%S}"),
                    European);
   exception
      when Occur : Tcl_Error_Exception =>
         Failed (Occur);
   end Property;

   -----------------------------
   -- Property (Long_Integer) --
   -----------------------------

   function Property (Query : Long_Property; Name : String) return Long_Integer is
      use Ada.Characters.Handling, Tcl.Ada;
   begin
      case Query is
         when Size =>
            return Long_Integer'Value (Eval ("file size " &
                                             Protected_Name (Name)));
      end case;
   exception
      when Occur : Tcl_Error_Exception =>
         Failed (Occur);
   end Property;

   ---------------------
   -- Property (Enum) --
   ---------------------

   Tcl_Enum_Info_Names : array (Enum_Property) of String (1..9) := (
      File_Kind => "type     ",
      Path_Kind => "pathtype ");

   function Property (Query : Enum_Property; Name : String) return Enum_Values is
      use Tcl.Ada;
   begin
      return Enum_Values'Value (Eval ("file " &
                                      Tcl_Enum_Info_Names(Query) &
                                      Protected_Name (Name)));
   exception
      when Occur : Tcl_Error_Exception =>
         Failed (Occur);
   end Property;


   ----------------------------------------------------------------------------------------
   --
   --  Directory iterator functions
   --
   ----------------------------------------------------------------------------------------

   ----------
   -- "+" --
   ----------

   function "+" (Left, Right : File_List) return File_List is
      use Tcl, Interfaces.C;
   begin
      -- Get rid of trivial case
      if Length(Left) = 0 then
         return Right;
      elsif Length(Right) = 0 then
         return Left;
      end if;

      -- Get rid of unsorted case
      if not Left.Content.Sorted or not Right.Content.Sorted then
         return (Ada.Finalization.Controlled with Content =>
            new Info_List'(Length    => Left.Content.Length + Right.Content.Length,
               Ref_Count => 1,
               Sorted    => False,
               Infos     => Left.Content.Infos & Right.Content.Infos)
            );
      end if;

      -- Here we are in the case of sorted "+"
      declare
         Left_Info  : Info_Table renames Left.Content.Infos;
         Right_Info : Info_Table renames Right.Content.Infos;
         Out_Info   : Info_Table (1..Left_Info'Length + Right_Info'Length); --Maximum possible size
         Left_Inx   : Natural := Left_Info'First;
         Right_Inx  : Natural := Right_Info'First;
         Out_Inx    : Natural := 0;
      begin
         while Left_Inx <= Left_Info'Last and Right_Inx <= Right_Info'Last loop
            case Dictionnary_Compare (Left_Info (Left_Inx).Dirname, Right_Info (Right_Inx).Dirname) is
            when Lower =>
               Out_Inx            := Out_Inx + 1;
               Out_Info (Out_Inx) := Left_Info (Left_Inx);
               Left_Inx           := Left_Inx + 1;
            when Greater =>
               Out_Inx            := Out_Inx + 1;
               Out_Info (Out_Inx) := Right_Info (Right_Inx);
               Right_Inx          := Right_Inx + 1;
            when Equal =>
               case Dictionnary_Compare (Left_Info (Left_Inx).Tail, Right_Info (Right_Inx).Tail) is
               when Lower =>
                  Out_Inx            := Out_Inx + 1;
                  Out_Info (Out_Inx) := Left_Info (Left_Inx);
                  Left_Inx           := Left_Inx + 1;
               when Greater =>
                  Out_Inx            := Out_Inx + 1;
                  Out_Info (Out_Inx) := Right_Info (Right_Inx);
                  Right_Inx          := Right_Inx + 1;
               when Equal => -- two identical entries
                  Out_Inx            := Out_Inx + 1;
                  Out_Info (Out_Inx) := Right_Info (Right_Inx);
                  Right_Inx          := Right_Inx + 1;
                  Left_Inx           := Left_Inx + 1;
               end case;
            end case;
         end loop;
         -- Add the non-exhausted side.
         -- The following assignments will do nothing for the exhausted side.
         Out_Info (Out_Inx+1 .. Out_Inx+1 + Left_Info'Last - Left_Inx) := Left_Info (Left_Inx .. Left_Info'Last);
         Out_Inx := Out_Inx + Left_Info'Last - Left_Inx + 1;
         Out_Info (Out_Inx+1 .. Out_Inx+1 + Right_Info'Last - Right_Inx) := Right_Info (Right_Inx .. Right_Info'Last);
         Out_Inx := Out_Inx + Right_Info'Last - Right_Inx + 1;

         return (Ada.Finalization.Controlled with Content =>
            new Info_List'(Length    => Out_Inx,
               Ref_Count => 1,
               Sorted    => True,
               Infos     => Out_Info (1..Out_Inx)
               )
            );
      end;
   end "+";

   ----------------
   -- Build_List --
   ----------------

   function Build_List (Pattern : in String  := "*";
                        Sort    : in Boolean := False)
   return File_List is
      use Tcl, Tcl.Ada, Ada.Finalization, Interfaces.C, Ada.Strings.Unbounded;

      L       : aliased Int;
      Result  : File_List;
      List    : constant Tcl_Obj := Eval("filelist {" & Protected_List (Pattern) & "} " & Boolean'Image (Sort));
      Elem    : aliased Tcl_Obj;

   begin -- Build_List
      Tcl_Incrrefcount (List);

      if Tcl_Listobjlength (Interp, List, L'access) /= Tcl_Ok then
         Tcl_Decrrefcount (List);
         Failed("Unable to get length in ""Build_List""");
      end if;

      Result.Content := new Info_List (Length => Natural (L));
      Result.Content.Ref_Count := 1;
      if L <= 1 then -- an empty list or a singleton is always sorted...
         Result.Content.Sorted := True;
      else
         Result.Content.Sorted := Sort;
      end if;
      for I in 0 .. L-1 loop
         Check_Result ( Tcl_ListObjIndex (Interp, List, I, Elem'access) );
         Tcl_Incrrefcount (Elem);
         declare
            Full_Name : String := Property (Expanded, Tcl_GetStringFromObj (Elem));
         begin
            Result.Content.Infos (Positive(I+1)) :=
               (Initialized => False,
                Tail        => To_Unbounded_String (Property (Name+Extension, Full_Name)),
                Dirname     => To_Unbounded_String (Property (Path,           Full_Name)));
         end;
         Tcl_Decrrefcount (Elem);
      end loop;

      Tcl_Decrrefcount (List);
      return Result;
   end Build_List;

   -----------------
   -- Check_Valid --
   -----------------

   -- Internal utility
   -- Check that the data in List(Index) is valid
   -- If not yet initialized, load it.

   procedure Check_Valid (List : File_List; Index : Natural) is
      function Get_Piece (Which : String) return String is
         use Tcl, Tcl.Ada;
      begin
         return Tcl_Getvar2 (Interp, "finfo", Which, Tcl_Global_Only);
      end Get_Piece;

      use Calendar_Utilities, Tcl.Ada;
   begin
      if not List.Content.Infos(Index).Initialized then
         Check_Result (Tcl_Eval (Interp, "fileinfo {" & Property (Expanded, List, Index) & '}'));
         List.Content.Infos(Index) :=
            (Initialized => True,
             Kind        => File_Kind_Enums'Value (Get_Piece ("type")),
             Tail        => List.Content.Infos(Index).Tail,
             Dirname     => List.Content.Infos(Index).Dirname,
             Atime       => Value (Eval ("clock format " & Get_Piece ("atime") & " -format {%d/%m/%Y %H:%M:%S}"),
                                   European),
             Mtime       => Value (Eval ("clock format " & Get_Piece ("mtime") & " -format {%d/%m/%Y %H:%M:%S}"),
                                   European),
            Size         => Long_Integer'Value (Get_Piece ("size"))
            );
      end if;
   end Check_Valid;

   --------------------
   -- Generic_Filter --
   --------------------

   function Generic_Filter (List : File_List) return File_List is
      use Ada.Finalization;

      Local_Table : Info_Table (1..List.Content.Infos'Length);
      Local_Inx   : Natural := 0;

   begin
      for I in 1..List.Content.Length loop
         if Is_Selected (List, I) then
            Local_Inx := Local_Inx+1;
            Local_Table (Local_Inx) := List.Content.Infos (I);
         end if;
      end loop;

      return (Controlled with new Info_List'(
            Length    => Local_Inx,
            Ref_Count => 1,
            Sorted    => List.Content.Sorted,
            Infos     => Local_Table (1..Local_Inx)));
   end Generic_Filter;

   ------------
   -- Length --
   ------------

   function Length (List : File_List) return Natural is
   begin
      if List.Content = null then
         return 0;
      else
         return List.Content.Length;
      end if;
   end Length;

   ----------------
   -- Properties --
   ----------------

   function  Properties (List : File_List; Index : Positive) return File_Properties is
      Elem : Inner_Info renames List.Content.Infos (Index);
   begin
      Check_Valid (List, Index);
      declare
         Result : File_Properties (Elem.Kind);
      begin
         Result.Name := Elem.Tail;
         Result.Path := Elem.Dirname;
         if Result.Kind in File..Directory then
            Result.Access_Time := Elem.Atime;
            Result.Change_Time := Elem.Mtime;
            Result.Size        := Elem.Size;
         end if;
         return Result;
      end;
   end Properties;

   -----------------------
   -- Property (string) --
   -----------------------

   function Property (Query : String_Property; List : File_List; Index : Positive) return String is
      -- This function deals only with *names*, which are always initialized.
      -- => do not call Check_Valid from this function
      --    (it is called FROM Check_Valid)
      use Ada.Strings.Unbounded;

      Elem : Inner_Info renames List.Content.Infos(Index);
   begin
      return Property (Query, Join (To_String (Elem.Dirname), To_String (Elem.Tail)));
   end Property;

   ---------------------
   -- Property (Enum) --
   ---------------------

   function Property (Query : Iter_Enum_Property; List : File_List; Index : Positive) return Enum_Values is
   -- Path_Type does not make sense, since we always provide the Path and Name.
   begin
      Check_Valid (List, Index);
      case Query is
         when File_Kind =>
            return List.Content.Infos(Index).Kind;
      end case;
   end Property;

   ------------------------
   -- Property (Boolean) --
   ------------------------

   function Property (Query : Boolean_Property; List : File_List; Index : Positive) return Boolean is
   begin
      case Query is
         when Is_Directory =>
            Check_Valid (List, Index);
            return List.Content.Infos(Index).Kind = Directory;
         when Is_Regular_File =>
            Check_Valid (List, Index);
            return List.Content.Infos(Index).Kind = File;
         when Exists | Is_Owned | Is_Readable | Is_Writable | Is_Executable =>
            return Property (Query, Property (Expanded, List, Index));
      end case;
   end Property;

   ---------------------
   -- Property (Time) --
   ---------------------

   function Property (Query : Time_Property;   List : File_List; Index : Positive) return Ada.Calendar.Time is
   begin
      Check_Valid (List, Index);
      case Query is
         when Access_Time =>
            return List.Content.Infos(Index).Atime;
         when Change_Time =>
            return List.Content.Infos(Index).Mtime;
      end case;
   end Property;

   -----------------------------
   -- Property (Long_Integer) --
   -----------------------------

   function Property (Query : Long_Property;   List : File_List; Index : Positive) return Long_Integer is
   begin
      Check_Valid (List, Index);
      case Query is
         when Size =>
            return List.Content.Infos(Index).Size;
      end case;
   end Property;

   ---------------
   -- To_String --
   ---------------

   function To_String (List : File_List; Separator : Character := ' ') return String is
      -- Note: there were several possibilities for writing this function.
      -- The one we chose has the benefit of being simple, and since this is a TCL implementation,
      -- we assume that efficiency is not too much a concern ;-).

      use Ada.Strings.Unbounded, Ada.Strings.Fixed;

      Result : Unbounded_String;
      String_Sep : constant String (1..1) := (1 => Separator);
   begin
      for Inx in 1..Length (List) loop
         declare
            File_Name : constant String := Property (Expanded, List, Inx);
         begin
            if Index (File_Name, String_Sep) = 0 then
               Result := Result & Separator & File_Name;
            else
               Result := Result & Separator & '"' & File_Name & '"';
            end if;
         end;
      end loop;

      -- Get rid of initial separator
      return To_String (Result)(2..Length (Result));
   end To_String;



   ----------------------------------------------------------------------------------------
   --
   --  File and directory operations
   --
   ----------------------------------------------------------------------------------------

   ----------------
   -- Change_Dir --
   ----------------

   procedure Change_Dir (To : String) is
      use Tcl.Ada;
   begin
      Tcl_Eval (Interp, "cd {" & To & '}');
   exception
      when Occur : Tcl_Error_Exception =>
         Failed (Occur);
   end Change_Dir;

   ---------------
   -- Copy_File --
   ---------------

   procedure Copy_File (Pattern : String; To : String) is
      use Tcl.Ada;
   begin
      Protected_Eval ("file copy -force -- " & Protected_List (Pattern) & " {" & To & '}');
   exception
      when Occur : Tcl_Error_Exception =>
         Failed (Occur);
   end Copy_File;

   ----------------
   -- Create_Dir --
   ----------------

   procedure Create_Dir (Name : String) is
      use Tcl.Ada;
   begin
      Protected_Eval ("file mkdir {" & Name & '}');
   exception
      when Occur : Tcl_Error_Exception =>
         Failed(Occur);
   end Create_Dir;

   -----------------
   -- Current_Dir --
   -----------------

   function Current_Dir return String is
      use Tcl.Ada;
   begin
      return Proper_Name (Eval ("pwd"), As_File => True);
   exception
      when Occur : Tcl_Error_Exception =>
         Failed (Occur);
   end Current_Dir;

   -----------------
   -- Delete_File --
   -----------------

   procedure Delete_File (Pattern : String) is
      use Tcl.Ada;
   begin
      Protected_Eval ("file delete -force -- " & Protected_List (Pattern));
   exception
      when Occur : Tcl_Error_Exception =>
         Failed (Occur);
   end Delete_File;

   ---------------
   -- Move_File --
   ---------------

   procedure Move_File (Pattern : String; To : String)
   is
   begin
      Copy_File (Pattern, To);
      Delete_File (Pattern); -- TBSL: if copy fails
   end Move_File;

   -----------------
   -- Rename_File --
   -----------------

   procedure Rename_File (File : String; To : String) is
      use Tcl.Ada;
   begin
      Protected_Eval ("file rename -force -- {" & File & "} {" & To & '}');
   exception
      when Occur : Tcl_Error_Exception =>
         Failed (Occur);
   end Rename_File;

   use Tcl.Ada;

begin
   -- Tcl says we must call this in order to use Tcl_GetNameOfExecutable
   Tcl_FindExecutable (Ada.Command_Line.Command_Name);

   Tcl_Eval (Interp,
      --tcl2ada OS_Services.tcl
--tcl2ada off
"proc fcomp {left right} {" & Ascii.LF &
"set res [string compare [string tolower $left] [string tolower $right]]" & Ascii.LF &
"if {$res == 0} {" & Ascii.LF &
"set res [string compare $left $right]" & Ascii.LF &
"}" & Ascii.LF &
"return $res" & Ascii.LF &
"}" & Ascii.LF &
"proc filelist {pattern sort} {" & Ascii.LF &
"set result [eval glob -nocomplain $pattern]" & Ascii.LF &
"if {$sort} {" & Ascii.LF &
"return [lsort -command fcomp $result]" & Ascii.LF &
"} else {" & Ascii.LF &
"return $result" & Ascii.LF &
"}" & Ascii.LF &
"}" & Ascii.LF &
"proc fileinfo {filename} {" & Ascii.LF &
"global finfo" & Ascii.LF &
"file stat $filename finfo" & Ascii.LF &
"}" & Ascii.LF &
""
--tcl2ada on
             );

   if not Forced_OS then
      -- Try to guess the OS
      declare
         OS_String : constant String := Tcl_Getvar(Interp, "tcl_platform(os)", 0);
      begin
         -- Strings according to the documentation. Tested only under Windows 95
         -- Don't know the string returned by OS/2. Is there a Tcl implementation?
         if OS_String = "Windows 95" then
            Current_OS := DOS;
         elsif OS_String = "Windows NT" then
            Current_OS := NT;
         elsif OS_String = "MacOS" then
            Current_OS := MacOS;
         else
            Current_OS := Unix;  -- Wild guess
         end if;
      end;
   end if;

   -- TBSL: I don't know for MacOS...
   Default_Ignore_Case := Current_OS /= Unix;

   case Operating_System is
      when Dos_OS =>
         Directory_Separator := '\';
      when Unix =>
         Directory_Separator := '/';
      when others =>
         Directory_Separator := ' '; --Unimplemented anyway
   end case;
end OS_Services_TCL;
